/*******************************************************************************
*                                                                              *
* foo_dsp_stereoconv                                                           *
* Copyright (C) 2007  Eric Larson (larson.eric.d@gmail.com)                    *
*                                                                              *
* This program is free software; you can redistribute it and/or modify         *
* it under the terms of the GNU General Public License as published by         *
* the Free Software Foundation; either version 2 of the License, or            *
* (at your option) any later version.                                          *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
*                                                                              *
* You should have received a copy of the GNU General Public License            *
* along with this program; if not, write to the Free Software                  *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA    *
*                                                                              *
*******************************************************************************/

#include "foo_dsp_stereoconv.h"
#include <windows.h>

static const GUID guid = { 0xf2d1f851, 0xba88, 0x4746, { 0x92, 0xbe, 0xd9, 0xbb, 0x5f, 0xcf, 0x87, 0x1d } };

DECLARE_COMPONENT_VERSION(
	"Stereo Convolution",
	"1.0.1",
	"foo_dsp_stereoconv\n"
	"Copyright 2007 Eric Larson\n"
	"http://people.bu.edu/edlarson/\n\n"
	);

class dialog_dsp_stereoconv : public dialog_helper::dialog_modal
{
public:
	dialog_dsp_stereoconv(t_dsp_stereoconv_params & p_params, dsp_preset_edit_callback & p_callback) : m_params(p_params), m_callback(p_callback) {	}

	virtual BOOL on_message(UINT msg, WPARAM wp, LPARAM lp)
	{
		switch(msg)
		{
		case WM_INITDIALOG:
			{
				HWND slider = GetDlgItem(get_wnd(), IDC_SLIDER);
				uSendMessage(slider, TBM_SETRANGE, TRUE, MAKELONG(-200, 200));
				uSendMessage(slider, TBM_SETPOS, TRUE, (int)(10*m_params.leveldB));
				uSetDlgItemText(get_wnd(), IDC_LEFTIN, m_params.lFile);
				uSetDlgItemText(get_wnd(), IDC_RIGHTIN, m_params.rFile);

				changed_since_saved = false;
				saved = false;
				
				update_fft_display();
				update_slider_display();
				update_button_display();

				// Set revert dsp settings MORE
				rev_params.leveldB = m_params.leveldB;
				strcpy_s(rev_params.lFile,NC,m_params.lFile);
				strcpy_s(rev_params.rFile,NC,m_params.rFile);
			}
			break;

			// Slider has been moved.
		case WM_HSCROLL:
			{
				changed_since_saved = true;
				saved = false;
				update_slider_display();
				update_button_display();
			}
			break;

		case WM_CLOSE:
			{
				// Data not changed
				end_dialog(0);
			}
			break;

		case WM_COMMAND:
			switch (LOWORD(wp))
			{
			case IDC_LEFTIN:  // Left Text
			case IDC_RIGHTIN: // Right Text
				if(HIWORD(wp) == EN_CHANGE) {
					changed_since_saved = true;
					update_button_display();
				}
			break;

			case IDC_BR_LEFT:  // Browse... left button
			case IDC_BR_RIGHT: // Browse... right button
				{

					OPENFILENAME ofn;       // common dialog box structure
					TCHAR szFile[NC];     // buffer for file name
					TCHAR filts[26] = L"Wave Files (*.wav)\0*.wav\0";
					HWND hwnd = get_wnd();  // owner window

					// Initialize OPENFILENAME
					ZeroMemory(&ofn, sizeof(ofn));
					ofn.lStructSize = sizeof(ofn);
					ofn.hwndOwner = hwnd;
					ofn.lpstrFile = szFile;
					ofn.lpstrFile[0] = '\0';
					ofn.nMaxFile = sizeof(szFile);
					ofn.lpstrFilter = filts;
					ofn.nFilterIndex = 1;
					ofn.lpstrFileTitle = NULL;
					ofn.nMaxFileTitle = 0;
					ofn.lpstrInitialDir = NULL;
					ofn.Flags = OFN_DONTADDTORECENT|OFN_NOCHANGEDIR;

					if(GetOpenFileName(&ofn)){
						char tempName[NC];
						wcstombs(tempName,ofn.lpstrFile,NC);

						if(wp == IDC_BR_LEFT){
							uSetDlgItemText(get_wnd(), IDC_LEFTIN, tempName);
						}
						else {
							uSetDlgItemText(get_wnd(), IDC_RIGHTIN, tempName);
						}
						changed_since_saved = true;
						saved = false;
						update_fft_display();
						update_button_display();
					}
				}
				break;

			case IDOK:
			case IDAPPLY:
				{
					if(changed_since_saved) {
						// Data has changed (again?), save it
						pfc::string8 tempPfc = pfc::string8("",256);
						m_params.leveldB = (double)uSendDlgItemMessage(get_wnd(), IDC_SLIDER, TBM_GETPOS, 0, 0)/10.0;
						uGetDlgItemText(get_wnd(), IDC_LEFTIN, tempPfc);
						strcpy_s(m_params.lFile,NC,tempPfc);
						uGetDlgItemText(get_wnd(), IDC_RIGHTIN, tempPfc);
						strcpy_s(m_params.rFile,NC,tempPfc);

						UpdatePreset();

						saved = true;
						changed_since_saved = false;

						update_button_display();
						update_fft_display();
					}
					if(wp == IDOK)
						end_dialog(1);
				}
				break;
			case IDREVERT:
			case IDCANCEL:
				{
					if(saved)
					{
						RevertSettings();
					}
					if(wp == IDCANCEL)
						end_dialog(0);
				}
				break;
			}
			break;
		}
		return 0;
	}

private:
	bool changed_since_saved;
	bool saved;
	t_dsp_stereoconv_params & m_params;
	t_dsp_stereoconv_params rev_params;
	dsp_preset_edit_callback & m_callback;

	static bool g_get_data(dsp_preset & p_data, t_dsp_stereoconv_params & tparams)
	{
		p_data.set_owner(guid);
		return tparams.get_data(p_data);
	}

	void UpdatePreset() {
		dsp_preset_impl temp;
		g_get_data(temp, m_params);
		m_callback.on_preset_changed(temp);
	}
	void RevertSettings() {
		m_params.leveldB = rev_params.leveldB;
		strcpy_s(m_params.lFile,NC,rev_params.lFile);
		strcpy_s(m_params.rFile,NC,rev_params.rFile);

		uSendMessage(GetDlgItem(get_wnd(), IDC_SLIDER), TBM_SETPOS, TRUE, (int)(10*m_params.leveldB));
		uSetDlgItemText(get_wnd(), IDC_LEFTIN, m_params.lFile);
		uSetDlgItemText(get_wnd(), IDC_RIGHTIN, m_params.rFile);
	
		changed_since_saved = false;
		saved = false;

		update_fft_display();
		update_slider_display();
		update_button_display();
		UpdatePreset();
	}
	void update_slider_display()
	{
		uSetDlgItemText(get_wnd(), IDC_STATIC_DISPLAY,
			pfc::string8() << pfc::format_float((float)uSendDlgItemMessage(get_wnd(), IDC_SLIDER, TBM_GETPOS, 0, 0)/10.0,0,1) << "dB");
	}


	void update_fft_display()
	{
		pfc::string8 tempPfc = pfc::string8("",NC);
		char tempLName[NC];
		char tempRName[NC];
		uGetDlgItemText(get_wnd(), IDC_LEFTIN, tempPfc); strcpy_s(tempLName,NC,tempPfc);
		uGetDlgItemText(get_wnd(), IDC_RIGHTIN, tempPfc); strcpy_s(tempRName,NC,tempPfc);

		dsp_abortback abort_cback;
		unsigned llen = 0;
		unsigned rlen = 0;
		get_wav_2ch(tempLName,&llen,NULL,NULL,NULL,abort_cback);
		get_wav_2ch(tempRName,&rlen,NULL,NULL,NULL,abort_cback);

		if(!llen)
			EnableWindow(GetDlgItem(get_wnd(),IDC_STATICL), false);
		else
			EnableWindow(GetDlgItem(get_wnd(),IDC_STATICL), true);
		if(!rlen)
			EnableWindow(GetDlgItem(get_wnd(),IDC_STATICR), false);
		else
			EnableWindow(GetDlgItem(get_wnd(),IDC_STATICR), true);

		uSetDlgItemText(get_wnd(), IDC_LENGTH,
			pfc::string8() << "FFT Length: " << pfc::format_int(GetWindowLen(llen,rlen)) << " samples");
	}

	void update_button_display() {
		if(!saved) {
			if(!changed_since_saved) { // not saved, not changed (no action taken)
				EnableWindow(GetDlgItem(get_wnd(),IDAPPLY), false);
				EnableWindow(GetDlgItem(get_wnd(),IDREVERT), false);
			}
			else{ // not saved but changed (modified)
				EnableWindow(GetDlgItem(get_wnd(),IDAPPLY), true);
				EnableWindow(GetDlgItem(get_wnd(),IDREVERT), false);
			}
		} else {
			if(!changed_since_saved) { // saved, not changed (modified, applied)
				EnableWindow(GetDlgItem(get_wnd(),IDAPPLY), false);
				EnableWindow(GetDlgItem(get_wnd(),IDREVERT), true);
			}
			else{ // saved and changed (modified, applied, modified)
				EnableWindow(GetDlgItem(get_wnd(),IDAPPLY), true);
				EnableWindow(GetDlgItem(get_wnd(),IDREVERT), true);
			}
		}
	}

};

class dsp_stereoconv : public dsp_impl_base {
private:
	int _sampleRate;
	int _bufferedSamples;

public:
	Convolve *_sc;
	t_dsp_stereoconv_params Params;

	dsp_stereoconv(const dsp_preset & p_data) {
		set_data(p_data);
		_sc = new Convolve(Params);
		_bufferedSamples = 0;
	}

	~dsp_stereoconv() {
		delete _sc;
	}

	static GUID g_get_guid() {
		return guid;
	}

	static void g_get_name(pfc::string_base & p_out) {
		p_out = "Stereo Convolver";
	}

	static bool g_have_config_popup() {
		return true;
	}

	static void g_show_config_popup(const dsp_preset & p_data, HWND p_parent, dsp_preset_edit_callback & p_callback) {
		t_dsp_stereoconv_params params;       // Create a params variable
		if (!params.set_data(p_data)) return; // Set params to the data in p_data
		dialog_dsp_stereoconv dlg(params,p_callback);    // Create a dialog using params
		dlg.run(IDD_CONFIG, p_parent);        // Run a dialog putting data into params
	}

	// Return the default preset
	static bool g_get_default_preset(dsp_preset & p_out) {
		t_dsp_stereoconv_params tparams;
		return g_get_data(p_out, tparams);
	}

	// Read parameters from a provided preset
	bool set_data(const dsp_preset & p_data)
	{
		if (!g_set_data(p_data, Params)) return false;
		return true;
	}

	// Helper method for reading parameters from a preset.
	static bool g_set_data(const dsp_preset & p_data, t_dsp_stereoconv_params & outparams) {
		if (p_data.get_owner() == g_get_guid())
		{
			t_dsp_stereoconv_params inparams;
			if (!inparams.set_data(p_data)) return false;
			outparams.leveldB = inparams.leveldB;
			strcpy_s(outparams.lFile,NC,inparams.lFile);
			strcpy_s(outparams.rFile,NC,inparams.rFile);
			return true;
		}
		return false;
	}

	// Helper method for writing parameters to a preset.
	static bool g_get_data(dsp_preset & p_data, t_dsp_stereoconv_params & tparams)
	{
		p_data.set_owner(g_get_guid());
		return tparams.get_data(p_data);
	}

	virtual bool on_chunk(audio_chunk * inChunk, abort_callback &) {
		_sampleRate = inChunk->get_sample_rate();
		if (inChunk->get_channels() != 2) {
			return true;
		}

		const audio_sample *inSamples = inChunk->get_data();
		int inSampleCount = (int)inChunk->get_sample_count();

		while (inSampleCount > 0) {
			int copyCount = min(inSampleCount, _sc->GetInputSamplesNeeded());
			double *scInput = _sc->GetInputBuffer();

			for (int i = 0; i < copyCount; i++) {
				scInput[0] = (double)inSamples[0];
				scInput[1] = (double)inSamples[1];
				scInput += 2;
				inSamples += 2;
			}
			inSampleCount -= copyCount;
			_bufferedSamples += copyCount;

			if (_sc->ProcessSamples(copyCount)) {
				OutputSamples(_sc->GetOutputBuffer(), _sc->GetOutputLength());
			}
		}

		return false;
	}

	virtual void flush() {
		_sc->Flush();
		_bufferedSamples = 0;
	}

	virtual void on_endoftrack(abort_callback &) {
		FlushToOutput();
	}

	virtual void on_endofplayback(abort_callback &) {
		FlushToOutput();
	}

	virtual double get_latency() {
		return ((double)_bufferedSamples / (double)_sampleRate);
	}

	virtual bool need_track_change_mark() {
		return false;
	}

private:
	void FlushToOutput() {
		int outSampleCount;

		while ((outSampleCount = _sc->FlushWithReturn()) != -1) {
			if (outSampleCount > 0) {
				OutputSamples(_sc->GetOutputBuffer(), outSampleCount);
			}
		}
	}

	void OutputSamples(double *scOutput, int outSampleCount) {
		audio_chunk *outChunk;
		audio_sample *outSamples;
		
		outChunk = insert_chunk(outSampleCount);
		outChunk->set_data_size(outSampleCount * 2);
		outChunk->set_sample_count(outSampleCount);
		outChunk->set_channels(2);
		outChunk->set_sample_rate(_sampleRate);
		outSamples = outChunk->get_data();

		for (int i = 0; i < outSampleCount; i++) {
			outSamples[0] = (audio_sample)scOutput[0];
			outSamples[1] = (audio_sample)scOutput[1];
			outSamples += 2;
			scOutput += 2;
		}

		_bufferedSamples -= outSampleCount;
	}

};

static dsp_factory_t<dsp_stereoconv> foo_dsp_stereoconv;